/*
 * Jalview - A Sequence Alignment Editor and Viewer ($$Version-Rel$$)
 * Copyright (C) $$Year-Rel$$ The Jalview Authors
 * 
 * This file is part of Jalview.
 * 
 * Jalview is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *  
 * Jalview is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty 
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jalview.  If not, see <http://www.gnu.org/licenses/>.
 * The Jalview Authors are detailed in the 'AUTHORS' file.
 */
package jalview.datamodel;

import jalview.util.Comparison;
import jalview.util.ShiftList;

import java.util.ArrayList;
import java.util.BitSet;
import java.util.Collections;
import java.util.List;
import java.util.Vector;
import java.util.concurrent.locks.ReentrantReadWriteLock;

public class HiddenColumns
{
  private static final ReentrantReadWriteLock LOCK = new ReentrantReadWriteLock();
  
  /*
   * list of hidden column [start, end] ranges; the list is maintained in
   * ascending start column order
   */
  private ArrayList<int[]> hiddenColumns;

  /**
   * Constructor
   */
  public HiddenColumns()
  {
  }

  /**
   * Copy constructor
   * 
   * @param copy
   */
  public HiddenColumns(HiddenColumns copy)
  {
    try
    {
      LOCK.writeLock().lock();
      if (copy != null)
      {
        if (copy.hiddenColumns != null)
        {
          hiddenColumns = copy.copyHiddenRegionsToArrayList();
        }
      }
    } finally
    {
      LOCK.writeLock().unlock();
    }
  }

  /**
   * This method is used to return all the HiddenColumn regions and is intended
   * to remain private. External callers which need a copy of the regions can
   * call getHiddenColumnsCopyAsList.
   * 
   * @return empty list or List of hidden column intervals
   */
  private List<int[]> getHiddenRegions()
  {
    return hiddenColumns == null ? Collections.<int[]> emptyList()
            : hiddenColumns;
  }

  /**
   * Output regions data as a string. String is in the format:
   * reg0[0]<between>reg0[1]<delimiter>reg1[0]<between>reg1[1] ... regn[1]
   * 
   * @param delimiter
   *          string to delimit regions
   * @param betweenstring
   *          to put between start and end region values
   * @return regions formatted according to delimiter and between strings
   */
  public String regionsToString(String delimiter, String between)
  {
    try
    {
      LOCK.readLock().lock();
      StringBuilder regionBuilder = new StringBuilder();
      if (hiddenColumns != null)
      {
        for (int[] range : hiddenColumns)
        {
          regionBuilder.append(delimiter).append(range[0]).append(between)
                  .append(range[1]);
        }

        regionBuilder.deleteCharAt(0);
      }
      return regionBuilder.toString();
    } finally
    {
      LOCK.readLock().unlock();
    }
  }

  /**
   * Find the number of hidden columns
   * 
   * @return number of hidden columns
   */
  public int getSize()
  {
    try
    {
      LOCK.readLock().lock();
      int size = 0;
      if (hasHiddenColumns())
      {
        for (int[] range : hiddenColumns)
        {
          size += range[1] - range[0] + 1;
        }
      }
      return size;
    }
    finally
    {
      LOCK.readLock().unlock();
    }
  }

  @Override
  public boolean equals(Object obj)
  {
    try
    {
      LOCK.readLock().lock();

      if (!(obj instanceof HiddenColumns))
      {
        return false;
      }
      HiddenColumns that = (HiddenColumns) obj;

      /*
       * check hidden columns are either both null, or match
       */
      if (this.hiddenColumns == null)
      {
        return (that.hiddenColumns == null);
      }
      if (that.hiddenColumns == null
              || that.hiddenColumns.size() != this.hiddenColumns.size())
      {
        return false;
      }
      int i = 0;
      for (int[] thisRange : hiddenColumns)
      {
        int[] thatRange = that.hiddenColumns.get(i++);
        if (thisRange[0] != thatRange[0] || thisRange[1] != thatRange[1])
        {
          return false;
        }
      }
      return true;
    } finally
    {
      LOCK.readLock().unlock();
    }
  }

  /**
   * Return absolute column index for a visible column index
   * 
   * @param column
   *          int column index in alignment view (count from zero)
   * @return alignment column index for column
   */
  public int adjustForHiddenColumns(int column)
  {
    try
    {
      LOCK.readLock().lock();
      int result = column;
      if (hiddenColumns != null)
      {
        for (int i = 0; i < hiddenColumns.size(); i++)
        {
          int[] region = hiddenColumns.get(i);
          if (result >= region[0])
          {
            result += region[1] - region[0] + 1;
          }
        }
      }
      return result;
    } finally
    {
      LOCK.readLock().unlock();
    }
  }

  /**
   * Use this method to find out where a column will appear in the visible
   * alignment when hidden columns exist. If the column is not visible, then the
   * left-most visible column will always be returned.
   * 
   * @param hiddenColumn
   *          the column index in the full alignment including hidden columns
   * @return the position of the column in the visible alignment
   */
  public int findColumnPosition(int hiddenColumn)
  {
    try
    {
      LOCK.readLock().lock();
      int result = hiddenColumn;
      if (hiddenColumns != null)
      {
        int index = 0;
        int[] region;
        do
        {
          region = hiddenColumns.get(index++);
          if (hiddenColumn > region[1])
          {
            result -= region[1] + 1 - region[0];
          }
        } while ((hiddenColumn > region[1])
                && (index < hiddenColumns.size()));

        if (hiddenColumn >= region[0] && hiddenColumn <= region[1])
        {
          // Here the hidden column is within a region, so
          // we want to return the position of region[0]-1, adjusted for any
          // earlier hidden columns.
          // Calculate the difference between the actual hidden col position
          // and region[0]-1, and then subtract from result to convert result
          // from
          // the adjusted hiddenColumn value to the adjusted region[0]-1 value

          // However, if the region begins at 0 we cannot return region[0]-1
          // just return 0
          if (region[0] == 0)
          {
            return 0;
          }
          else
          {
            return result - (hiddenColumn - region[0] + 1);
          }
        }
      }
      return result; // return the shifted position after removing hidden
                     // columns.
    } finally
    {
      LOCK.readLock().unlock();
    }
  }

  /**
   * Find the visible column which is a given visible number of columns to the
   * left of another visible column. i.e. for a startColumn x, the column which
   * is distance 1 away will be column x-1.
   * 
   * @param visibleDistance
   *          the number of visible columns to offset by
   * @param startColumn
   *          the column to start from
   * @return the position of the column in the visible alignment
   */
  public int subtractVisibleColumns(int visibleDistance, int startColumn)
  {
    try
    {

      LOCK.readLock().lock();
    int distance = visibleDistance;

    // in case startColumn is in a hidden region, move it to the left
    int start = adjustForHiddenColumns(findColumnPosition(startColumn));

    // get index of hidden region to left of start
    int index = getHiddenIndexLeft(start);
    if (index == -1)
    {
      // no hidden regions to left of startColumn
      return start - distance;
    }

    // walk backwards through the alignment subtracting the counts of visible
    // columns from distance
    int[] region;
    int gap = 0;
    int nextstart = start;

    while ((index > -1) && (distance - gap > 0))
    {
      // subtract the gap to right of region from distance
      distance -= gap;
      start = nextstart;

      // calculate the next gap
      region = hiddenColumns.get(index);
      gap = start - region[1];

      // set start to just to left of current region
      nextstart = region[0] - 1;
      index--;
    }

    if (distance - gap > 0)
    {
      // fell out of loop because there are no more hidden regions
      distance -= gap;
      return nextstart - distance;
    }
    return start - distance;
    } finally
    {
      LOCK.readLock().unlock();
    }

  }

  /**
   * Use this method to determine the set of hiddenRegion start positions
   * 
   * @return list of column number in visible view where hidden regions start
   */
  public List<Integer> findHiddenRegionPositions()
  {
    try
    {
      LOCK.readLock().lock();
      List<Integer> positions = null;

      if (hiddenColumns != null)
      {
        positions = new ArrayList<>(hiddenColumns.size());

        positions.add(hiddenColumns.get(0)[0]);
        for (int i = 1; i < hiddenColumns.size(); ++i)
        {

          int result = 0;
          if (hiddenColumns != null)
          {
            int index = 0;
            int gaps = 0;
            do
            {
              int[] region = hiddenColumns.get(index);
              gaps += region[1] + 1 - region[0];
              result = region[1] + 1;
              index++;
            } while (index <= i);

            result -= gaps;
          }
          positions.add(result);
        }
      }
      else
      {
        positions = new ArrayList<>();
      }

      return positions;
    }
    finally
    {
      LOCK.readLock().unlock();
    }
  }

  /**
   * This method returns the rightmost limit of a region of an alignment with
   * hidden columns. In otherwords, the next hidden column.
   * 
   * @param index
   *          int
   */
  public int getHiddenBoundaryRight(int alPos)
  {
    try
    {
      LOCK.readLock().lock();
      if (hiddenColumns != null)
      {
        int index = 0;
        do
        {
          int[] region = hiddenColumns.get(index);
          if (alPos < region[0])
          {
            return region[0];
          }

          index++;
        } while (index < hiddenColumns.size());
      }

      return alPos;
    } finally
    {
      LOCK.readLock().unlock();
    }

  }

  /**
   * This method returns the leftmost limit of a region of an alignment with
   * hidden columns. In otherwords, the previous hidden column.
   * 
   * @param index
   *          int
   */
  public int getHiddenBoundaryLeft(int alPos)
  {
    try
    {
      LOCK.readLock().lock();

    if (hiddenColumns != null)
    {
      int index = hiddenColumns.size() - 1;
      do
      {
          int[] region = hiddenColumns.get(index);
        if (alPos > region[1])
        {
          return region[1];
        }

        index--;
      } while (index > -1);
    }

    return alPos;
    } finally
    {
      LOCK.readLock().unlock();
    }
  }

  /**
   * This method returns the index of the hidden region to the left of a column
   * position. If the column is in a hidden region it returns the index of the
   * region to the left. If there is no hidden region to the left it returns -1.
   * 
   * @param pos
   *          int
   */
  private int getHiddenIndexLeft(int pos)
  {
    try
    {

      LOCK.readLock().lock();
    if (hiddenColumns != null)
    {
      int index = hiddenColumns.size() - 1;
      do
      {
          int[] region = hiddenColumns.get(index);
        if (pos > region[1])
        {
          return index;
        }

        index--;
      } while (index > -1);
    }

    return -1;
    } finally
    {
      LOCK.readLock().unlock();
    }

  }

  /**
   * Adds the specified column range to the hidden columns
   * 
   * @param start
   * @param end
   */
  public void hideColumns(int start, int end)
  {
    boolean wasAlreadyLocked = false;
    try
    {
      // check if the write lock was already locked by this thread,
      // as this method can be called internally in loops within HiddenColumns
      if (!LOCK.isWriteLockedByCurrentThread())
      {
        LOCK.writeLock().lock();
      }
      else
      {
        wasAlreadyLocked = true;
      }

      if (hiddenColumns == null)
      {
        hiddenColumns = new ArrayList<>();
      }

      /*
       * traverse existing hidden ranges and insert / amend / append as
       * appropriate
       */
      for (int i = 0; i < hiddenColumns.size(); i++)
      {
        int[] region = hiddenColumns.get(i);

        if (end < region[0] - 1)
        {
          /*
           * insert discontiguous preceding range
           */
          hiddenColumns.add(i, new int[] { start, end });
          return;
        }

        if (end <= region[1])
        {
          /*
           * new range overlaps existing, or is contiguous preceding it - adjust
           * start column
           */
          region[0] = Math.min(region[0], start);
          return;
        }

        if (start <= region[1] + 1)
        {
          /*
           * new range overlaps existing, or is contiguous following it - adjust
           * start and end columns
           */
          region[0] = Math.min(region[0], start);
          region[1] = Math.max(region[1], end);

          /*
           * also update or remove any subsequent ranges 
           * that are overlapped
           */
          while (i < hiddenColumns.size() - 1)
          {
            int[] nextRegion = hiddenColumns.get(i + 1);
            if (nextRegion[0] > end + 1)
            {
              /*
               * gap to next hidden range - no more to update
               */
              break;
            }
            region[1] = Math.max(nextRegion[1], end);
            hiddenColumns.remove(i + 1);
          }
          return;
        }
    }

    /*
     * remaining case is that the new range follows everything else
     */
      hiddenColumns.add(new int[] { start, end });
    } finally
    {
      if (!wasAlreadyLocked)
      {
        LOCK.writeLock().unlock();
      }
    }
  }

  public boolean isVisible(int column)
  {
    try
    {
      LOCK.readLock().lock();

    if (hiddenColumns != null)
    {
      for (int[] region : hiddenColumns)
      {
        if (column >= region[0] && column <= region[1])
        {
          return false;
        }
      }
    }

    return true;
    } finally
    {
      LOCK.readLock().unlock();
    }
  }

  private ArrayList<int[]> copyHiddenRegionsToArrayList()
  {
    int size = 0;
    if (hiddenColumns != null)
    {
      size = hiddenColumns.size();
    }
    ArrayList<int[]> copy = new ArrayList<>(size);

    for (int i = 0, j = size; i < j; i++)
    {
      int[] rh;
      int[] cp;
      rh = hiddenColumns.get(i);
      if (rh != null)
      {
        cp = new int[rh.length];
        System.arraycopy(rh, 0, cp, 0, rh.length);
        copy.add(cp);
      }
    }

    return copy;
  }
  
  /**
   * Returns a copy of the vector of hidden regions, as an ArrayList. Before
   * using this method please consider if you really need access to the hidden
   * regions - a new (or existing!) method on HiddenColumns might be more
   * appropriate.
   * 
   * @return hidden regions as an ArrayList of [start,end] pairs
   */
  public ArrayList<int[]> getHiddenColumnsCopy()
  {
    try
    {
      LOCK.readLock().lock();
      return copyHiddenRegionsToArrayList();
    } finally
    {
      LOCK.readLock().unlock();
    }
  }

  /**
   * propagate shift in alignment columns to column selection
   * 
   * @param start
   *          beginning of edit
   * @param left
   *          shift in edit (+ve for removal, or -ve for inserts)
   */
  public List<int[]> compensateForEdit(int start, int change,
          ColumnSelection sel)
  {
    try
    {
      LOCK.writeLock().lock();
      List<int[]> deletedHiddenColumns = null;

      if (hiddenColumns != null)
      {
        deletedHiddenColumns = new ArrayList<>();
        int hSize = hiddenColumns.size();
        for (int i = 0; i < hSize; i++)
        {
          int[] region = hiddenColumns.get(i);
          if (region[0] > start && start + change > region[1])
          {
            deletedHiddenColumns.add(region);

            hiddenColumns.remove(i);
            i--;
            hSize--;
            continue;
          }

          if (region[0] > start)
          {
            region[0] -= change;
            region[1] -= change;
          }

          if (region[0] < 0)
          {
            region[0] = 0;
          }

        }

        this.revealHiddenColumns(0, sel);
      }

      return deletedHiddenColumns;
    } finally
    {
      LOCK.writeLock().unlock();
    }
  }

  /**
   * propagate shift in alignment columns to column selection special version of
   * compensateForEdit - allowing for edits within hidden regions
   * 
   * @param start
   *          beginning of edit
   * @param left
   *          shift in edit (+ve for removal, or -ve for inserts)
   */
  public void compensateForDelEdits(int start, int change)
  {
    try
    {
      LOCK.writeLock().lock();
      if (hiddenColumns != null)
      {
        for (int i = 0; i < hiddenColumns.size(); i++)
        {
          int[] region = hiddenColumns.get(i);
          if (region[0] >= start)
          {
            region[0] -= change;
          }
          if (region[1] >= start)
          {
            region[1] -= change;
          }
          if (region[1] < region[0])
          {
            hiddenColumns.remove(i--);
          }

          if (region[0] < 0)
          {
            region[0] = 0;
          }
          if (region[1] < 0)
          {
            region[1] = 0;
          }
        }
      }
    }
    finally
    {
      LOCK.writeLock().unlock();
    }
  }

  /**
   * return all visible segments between the given start and end boundaries
   * 
   * @param start
   *          (first column inclusive from 0)
   * @param end
   *          (last column - not inclusive)
   * @return int[] {i_start, i_end, ..} where intervals lie in
   *         start<=i_start<=i_end<end
   */
  public int[] getVisibleContigs(int start, int end)
  {
    try
    {
      LOCK.readLock().lock();
      if (hiddenColumns != null && hiddenColumns.size() > 0)
      {
        List<int[]> visiblecontigs = new ArrayList<>();
        List<int[]> regions = getHiddenRegions();

        int vstart = start;
        int[] region;
        int hideStart;
        int hideEnd;

        for (int j = 0; vstart < end && j < regions.size(); j++)
        {
          region = regions.get(j);
          hideStart = region[0];
          hideEnd = region[1];

          if (hideEnd < vstart)
          {
            continue;
          }
          if (hideStart > vstart)
          {
            visiblecontigs.add(new int[] { vstart, hideStart - 1 });
          }
          vstart = hideEnd + 1;
        }

        if (vstart < end)
        {
          visiblecontigs.add(new int[] { vstart, end - 1 });
        }
        int[] vcontigs = new int[visiblecontigs.size() * 2];
        for (int i = 0, j = visiblecontigs.size(); i < j; i++)
        {
          int[] vc = visiblecontigs.get(i);
          visiblecontigs.set(i, null);
          vcontigs[i * 2] = vc[0];
          vcontigs[i * 2 + 1] = vc[1];
        }
        visiblecontigs.clear();
        return vcontigs;
      }
      else
      {
        return new int[] { start, end - 1 };
      }
    }
    finally
    {
      LOCK.readLock().unlock();
    }
  }

  public String[] getVisibleSequenceStrings(int start, int end,
          SequenceI[] seqs)
  {
    try
    {
      LOCK.readLock().lock();
      int iSize = seqs.length;
      String[] selections = new String[iSize];
      if (hiddenColumns != null && hiddenColumns.size() > 0)
      {
        for (int i = 0; i < iSize; i++)
        {
          StringBuffer visibleSeq = new StringBuffer();
          List<int[]> regions = getHiddenRegions();

          int blockStart = start;
          int blockEnd = end;
          int[] region;
          int hideStart;
          int hideEnd;

          for (int j = 0; j < regions.size(); j++)
          {
            region = regions.get(j);
            hideStart = region[0];
            hideEnd = region[1];

            if (hideStart < start)
            {
              continue;
            }

            blockStart = Math.min(blockStart, hideEnd + 1);
            blockEnd = Math.min(blockEnd, hideStart);

            if (blockStart > blockEnd)
            {
              break;
            }

            visibleSeq.append(seqs[i].getSequence(blockStart, blockEnd));

            blockStart = hideEnd + 1;
            blockEnd = end;
          }

          if (end > blockStart)
          {
            visibleSeq.append(seqs[i].getSequence(blockStart, end));
          }

          selections[i] = visibleSeq.toString();
        }
      }
      else
      {
        for (int i = 0; i < iSize; i++)
        {
          selections[i] = seqs[i].getSequenceAsString(start, end);
        }
      }

      return selections;
    }
    finally
    {
      LOCK.readLock().unlock();
    }
  }

  /**
   * Locate the first and last position visible for this sequence. if seq isn't
   * visible then return the position of the left and right of the hidden
   * boundary region, and the corresponding alignment column indices for the
   * extent of the sequence
   * 
   * @param seq
   * @return int[] { visible start, visible end, first seqpos, last seqpos,
   *         alignment index for seq start, alignment index for seq end }
   */
  public int[] locateVisibleBoundsOfSequence(SequenceI seq)
  {
    try
    {
      LOCK.readLock().lock();
      int fpos = seq.getStart();
      int lpos = seq.getEnd();
      int start = 0;

      if (hiddenColumns == null || hiddenColumns.size() == 0)
      {
        int ifpos = seq.findIndex(fpos) - 1;
        int ilpos = seq.findIndex(lpos) - 1;
        return new int[] { ifpos, ilpos, fpos, lpos, ifpos, ilpos };
      }

      // Simply walk along the sequence whilst watching for hidden column
      // boundaries
      List<int[]> regions = getHiddenRegions();
      int spos = fpos;
      int lastvispos = -1;
      int rcount = 0;
      int hideStart = seq.getLength();
      int hideEnd = -1;
      int visPrev = 0;
      int visNext = 0;
      int firstP = -1;
      int lastP = -1;
      boolean foundStart = false;
      for (int p = 0, pLen = seq.getLength(); spos <= seq.getEnd()
              && p < pLen; p++)
      {
        if (!Comparison.isGap(seq.getCharAt(p)))
        {
          // keep track of first/last column
          // containing sequence data regardless of visibility
          if (firstP == -1)
          {
            firstP = p;
          }
          lastP = p;
          // update hidden region start/end
          while (hideEnd < p && rcount < regions.size())
          {
            int[] region = regions.get(rcount++);
            visPrev = visNext;
            visNext += region[0] - visPrev;
            hideStart = region[0];
            hideEnd = region[1];
          }
          if (hideEnd < p)
          {
            hideStart = seq.getLength();
          }
          // update visible boundary for sequence
          if (p < hideStart)
          {
            if (!foundStart)
            {
              fpos = spos;
              start = p;
              foundStart = true;
            }
            lastvispos = p;
            lpos = spos;
          }
          // look for next sequence position
          spos++;
        }
      }
      if (foundStart)
      {
        return new int[] { findColumnPosition(start),
            findColumnPosition(lastvispos), fpos, lpos, firstP, lastP };
      }
      // otherwise, sequence was completely hidden
      return new int[] { visPrev, visNext, 0, 0, firstP, lastP };
    }
    finally
    {
      LOCK.readLock().unlock();
    }
  }

  /**
   * delete any columns in alignmentAnnotation that are hidden (including
   * sequence associated annotation).
   * 
   * @param alignmentAnnotation
   */
  public void makeVisibleAnnotation(AlignmentAnnotation alignmentAnnotation)
  {
    makeVisibleAnnotation(-1, -1, alignmentAnnotation);
  }

  /**
   * delete any columns in alignmentAnnotation that are hidden (including
   * sequence associated annotation).
   * 
   * @param start
   *          remove any annotation to the right of this column
   * @param end
   *          remove any annotation to the left of this column
   * @param alignmentAnnotation
   *          the annotation to operate on
   */
  public void makeVisibleAnnotation(int start, int end,
          AlignmentAnnotation alignmentAnnotation)
  {
    try
    {
      LOCK.readLock().lock();
      if (alignmentAnnotation.annotations == null)
      {
        return;
      }
      if (start == end && end == -1)
      {
        start = 0;
        end = alignmentAnnotation.annotations.length;
      }
      if (hiddenColumns != null && hiddenColumns.size() > 0)
      {
        // then mangle the alignmentAnnotation annotation array
        Vector<Annotation[]> annels = new Vector<>();
        Annotation[] els = null;
        List<int[]> regions = getHiddenRegions();
        int blockStart = start;
        int blockEnd = end;
        int[] region;
        int hideStart;
        int hideEnd;
        int w = 0;

        for (int j = 0; j < regions.size(); j++)
        {
          region = regions.get(j);
          hideStart = region[0];
          hideEnd = region[1];

          if (hideStart < start)
          {
            continue;
          }

          blockStart = Math.min(blockStart, hideEnd + 1);
          blockEnd = Math.min(blockEnd, hideStart);

          if (blockStart > blockEnd)
          {
            break;
          }

          annels.addElement(els = new Annotation[blockEnd - blockStart]);
          System.arraycopy(alignmentAnnotation.annotations, blockStart, els,
                  0, els.length);
          w += els.length;
          blockStart = hideEnd + 1;
          blockEnd = end;
        }

        if (end > blockStart)
        {
          annels.addElement(els = new Annotation[end - blockStart + 1]);
          if ((els.length
                  + blockStart) <= alignmentAnnotation.annotations.length)
          {
            // copy just the visible segment of the annotation row
            System.arraycopy(alignmentAnnotation.annotations, blockStart,
                    els, 0, els.length);
          }
          else
          {
            // copy to the end of the annotation row
            System.arraycopy(alignmentAnnotation.annotations, blockStart,
                    els, 0,
                    (alignmentAnnotation.annotations.length - blockStart));
          }
          w += els.length;
        }
        if (w == 0)
        {
          return;
        }

        alignmentAnnotation.annotations = new Annotation[w];
        w = 0;

        for (Annotation[] chnk : annels)
        {
          System.arraycopy(chnk, 0, alignmentAnnotation.annotations, w,
                  chnk.length);
          w += chnk.length;
        }
      }
      else
      {
        alignmentAnnotation.restrict(start, end);
      }
    }
    finally
    {
      LOCK.readLock().unlock();
    }
  }

  /**
   * 
   * @return true if there are columns hidden
   */
  public boolean hasHiddenColumns()
  {
    try
    {
      LOCK.readLock().lock();
      return hiddenColumns != null && hiddenColumns.size() > 0;
    } finally
    {
      LOCK.readLock().unlock();
    }
  }

  /**
   * 
   * @return true if there are more than one set of columns hidden
   */
  public boolean hasManyHiddenColumns()
  {
    try
    {
      LOCK.readLock().lock();
      return hiddenColumns != null && hiddenColumns.size() > 1;
    } finally
    {
      LOCK.readLock().unlock();
    }
  }

  /**
   * mark the columns corresponding to gap characters as hidden in the column
   * selection
   * 
   * @param sr
   */
  public void hideInsertionsFor(SequenceI sr)
  {
    try
    {
      LOCK.writeLock().lock();
      List<int[]> inserts = sr.getInsertions();
      for (int[] r : inserts)
      {
        hideColumns(r[0], r[1]);
      }
    } finally
    {
      LOCK.writeLock().unlock();
    }
  }

  /**
   * Unhides, and adds to the selection list, all hidden columns
   */
  public void revealAllHiddenColumns(ColumnSelection sel)
  {
    try
    {
      LOCK.writeLock().lock();
      if (hiddenColumns != null)
      {
        for (int i = 0; i < hiddenColumns.size(); i++)
        {
          int[] region = hiddenColumns.get(i);
          for (int j = region[0]; j < region[1] + 1; j++)
          {
            sel.addElement(j);
          }
        }
      }

      hiddenColumns = null;
    }
    finally
    {
      LOCK.writeLock().unlock();
    }
  }

  /**
   * Reveals, and marks as selected, the hidden column range with the given
   * start column
   * 
   * @param start
   */
  public void revealHiddenColumns(int start, ColumnSelection sel)
  {
    try
    {
      LOCK.writeLock().lock();
      for (int i = 0; i < hiddenColumns.size(); i++)
      {
        int[] region = hiddenColumns.get(i);
        if (start == region[0])
        {
          for (int j = region[0]; j < region[1] + 1; j++)
          {
            sel.addElement(j);
          }

          hiddenColumns.remove(region);
          break;
        }
      }
      if (hiddenColumns.size() == 0)
      {
        hiddenColumns = null;
      }
    }
    finally
    {
      LOCK.writeLock().unlock();
    }
  }

  /**
   * removes intersection of position,length ranges in deletions from the
   * start,end regions marked in intervals.
   * 
   * @param shifts
   * @param intervals
   * @return
   */
  private boolean pruneIntervalList(final List<int[]> shifts,
          ArrayList<int[]> intervals)
  {
    boolean pruned = false;
    int i = 0;
    int j = intervals.size() - 1;
    int s = 0;
    int t = shifts.size() - 1;
    int[] hr = intervals.get(i);
    int[] sr = shifts.get(s);
    while (i <= j && s <= t)
    {
      boolean trailinghn = hr[1] >= sr[0];
      if (!trailinghn)
      {
        if (i < j)
        {
          hr = intervals.get(++i);
        }
        else
        {
          i++;
        }
        continue;
      }
      int endshift = sr[0] + sr[1]; // deletion ranges - -ve means an insert
      if (endshift < hr[0] || endshift < sr[0])
      { // leadinghc disjoint or not a deletion
        if (s < t)
        {
          sr = shifts.get(++s);
        }
        else
        {
          s++;
        }
        continue;
      }
      boolean leadinghn = hr[0] >= sr[0];
      boolean leadinghc = hr[0] < endshift;
      boolean trailinghc = hr[1] < endshift;
      if (leadinghn)
      {
        if (trailinghc)
        { // deleted hidden region.
          intervals.remove(i);
          pruned = true;
          j--;
          if (i <= j)
          {
            hr = intervals.get(i);
          }
          continue;
        }
        if (leadinghc)
        {
          hr[0] = endshift; // clip c terminal region
          leadinghn = !leadinghn;
          pruned = true;
        }
      }
      if (!leadinghn)
      {
        if (trailinghc)
        {
          if (trailinghn)
          {
            hr[1] = sr[0] - 1;
            pruned = true;
          }
        }
        else
        {
          // sr contained in hr
          if (s < t)
          {
            sr = shifts.get(++s);
          }
          else
          {
            s++;
          }
          continue;
        }
      }
    }
    return pruned; // true if any interval was removed or modified by
    // operations.
  }

  /**
   * remove any hiddenColumns or selected columns and shift remaining based on a
   * series of position, range deletions.
   * 
   * @param deletions
   */
  public void pruneDeletions(List<int[]> shifts)
  {
    try
    {
      LOCK.writeLock().lock();
      // delete any intervals intersecting.
      if (hiddenColumns != null)
      {
        pruneIntervalList(shifts, hiddenColumns);
        if (hiddenColumns != null && hiddenColumns.size() == 0)
        {
          hiddenColumns = null;
        }
      }
    }
    finally
    {
      LOCK.writeLock().unlock();
    }
  }

  /**
   * Add gaps into the sequences aligned to profileseq under the given
   * AlignmentView
   * 
   * @param profileseq
   * @param al
   *          - alignment to have gaps inserted into it
   * @param input
   *          - alignment view where sequence corresponding to profileseq is
   *          first entry
   * @return new HiddenColumns for new alignment view, with insertions into
   *         profileseq marked as hidden.
   */
  public static HiddenColumns propagateInsertions(SequenceI profileseq,
          AlignmentI al, AlignmentView input)
  {
    int profsqpos = 0;

    char gc = al.getGapCharacter();
    Object[] alandhidden = input.getAlignmentAndHiddenColumns(gc);
    HiddenColumns nview = (HiddenColumns) alandhidden[1];
    SequenceI origseq = ((SequenceI[]) alandhidden[0])[profsqpos];
    nview.propagateInsertions(profileseq, al, origseq);
    return nview;
  }

  /**
   * 
   * @param profileseq
   *          - sequence in al which corresponds to origseq
   * @param al
   *          - alignment which is to have gaps inserted into it
   * @param origseq
   *          - sequence corresponding to profileseq which defines gap map for
   *          modifying al
   */
  private void propagateInsertions(SequenceI profileseq, AlignmentI al,
          SequenceI origseq)
  {
    char gc = al.getGapCharacter();
    // recover mapping between sequence's non-gap positions and positions
    // mapping to view.
    pruneDeletions(ShiftList.parseMap(origseq.gapMap()));
    int[] viscontigs = getVisibleContigs(0, profileseq.getLength());
    int spos = 0;
    int offset = 0;

    // add profile to visible contigs
    for (int v = 0; v < viscontigs.length; v += 2)
    {
      if (viscontigs[v] > spos)
      {
        StringBuffer sb = new StringBuffer();
        for (int s = 0, ns = viscontigs[v] - spos; s < ns; s++)
        {
          sb.append(gc);
        }
        for (int s = 0, ns = al.getHeight(); s < ns; s++)
        {
          SequenceI sqobj = al.getSequenceAt(s);
          if (sqobj != profileseq)
          {
            String sq = al.getSequenceAt(s).getSequenceAsString();
            if (sq.length() <= spos + offset)
            {
              // pad sequence
              int diff = spos + offset - sq.length() - 1;
              if (diff > 0)
              {
                // pad gaps
                sq = sq + sb;
                while ((diff = spos + offset - sq.length() - 1) > 0)
                {
                  // sq = sq
                  // + ((diff >= sb.length()) ? sb.toString() : sb
                  // .substring(0, diff));
                  if (diff >= sb.length())
                  {
                    sq += sb.toString();
                  }
                  else
                  {
                    char[] buf = new char[diff];
                    sb.getChars(0, diff, buf, 0);
                    sq += buf.toString();
                  }
                }
              }
              sq += sb.toString();
            }
            else
            {
              al.getSequenceAt(s).setSequence(
                      sq.substring(0, spos + offset) + sb.toString()
                              + sq.substring(spos + offset));
            }
          }
        }
        // offset+=sb.length();
      }
      spos = viscontigs[v + 1] + 1;
    }
    if ((offset + spos) < profileseq.getLength())
    {
      // pad the final region with gaps.
      StringBuffer sb = new StringBuffer();
      for (int s = 0, ns = profileseq.getLength() - spos - offset; s < ns; s++)
      {
        sb.append(gc);
      }
      for (int s = 0, ns = al.getHeight(); s < ns; s++)
      {
        SequenceI sqobj = al.getSequenceAt(s);
        if (sqobj == profileseq)
        {
          continue;
        }
        String sq = sqobj.getSequenceAsString();
        // pad sequence
        int diff = origseq.getLength() - sq.length();
        while (diff > 0)
        {
          // sq = sq
          // + ((diff >= sb.length()) ? sb.toString() : sb
          // .substring(0, diff));
          if (diff >= sb.length())
          {
            sq += sb.toString();
          }
          else
          {
            char[] buf = new char[diff];
            sb.getChars(0, diff, buf, 0);
            sq += buf.toString();
          }
          diff = origseq.getLength() - sq.length();
        }
      }
    }
  }

  /**
   * remove any hiddenColumns or selected columns and shift remaining based on a
   * series of position, range deletions.
   * 
   * @param deletions
   */
  private void pruneDeletions(ShiftList deletions)
  {
    if (deletions != null)
    {
      final List<int[]> shifts = deletions.getShifts();
      if (shifts != null && shifts.size() > 0)
      {
        pruneDeletions(shifts);

        // and shift the rest.
        this.compensateForEdits(deletions);
      }
    }
  }

  /**
   * Adjust hidden column boundaries based on a series of column additions or
   * deletions in visible regions.
   * 
   * @param shiftrecord
   * @return
   */
  private ShiftList compensateForEdits(ShiftList shiftrecord)
  {
    if (shiftrecord != null)
    {
      final List<int[]> shifts = shiftrecord.getShifts();
      if (shifts != null && shifts.size() > 0)
      {
        int shifted = 0;
        for (int i = 0, j = shifts.size(); i < j; i++)
        {
          int[] sh = shifts.get(i);
          compensateForDelEdits(shifted + sh[0], sh[1]);
          shifted -= sh[1];
        }
      }
      return shiftrecord.getInverse();
    }
    return null;
  }

  /**
   * Returns a hashCode built from hidden column ranges
   */
  @Override
  public int hashCode()
  {
    try
    {
      LOCK.readLock().lock();
      int hashCode = 1;
      if (hiddenColumns != null)
      {
        for (int[] hidden : hiddenColumns)
        {
          hashCode = 31 * hashCode + hidden[0];
          hashCode = 31 * hashCode + hidden[1];
        }
      }
      return hashCode;
    }
    finally
    {
      LOCK.readLock().unlock();
    }
  }

  /**
   * Hide columns corresponding to the marked bits
   * 
   * @param inserts
   *          - columns map to bits starting from zero
   */
  public void hideMarkedBits(BitSet inserts)
  {
    try
    {
      LOCK.writeLock().lock();
      for (int firstSet = inserts
              .nextSetBit(0), lastSet = 0; firstSet >= 0; firstSet = inserts
                      .nextSetBit(lastSet))
      {
        lastSet = inserts.nextClearBit(firstSet);
        hideColumns(firstSet, lastSet - 1);
      }
    } finally
    {
      LOCK.writeLock().unlock();
    }
  }

  /**
   * 
   * @param inserts
   *          BitSet where hidden columns will be marked
   */
  public void markHiddenRegions(BitSet inserts)
  {
    try
    {
      LOCK.readLock().lock();
      if (hiddenColumns == null)
      {
        return;
      }
      for (int[] range : hiddenColumns)
      {
        inserts.set(range[0], range[1] + 1);
      }
    }
    finally
    {
      LOCK.readLock().unlock();
    }
  }

  /**
   * Calculate the visible start and end index of an alignment.
   * 
   * @param width
   *          full alignment width
   * @return integer array where: int[0] = startIndex, and int[1] = endIndex
   */
  public int[] getVisibleStartAndEndIndex(int width)
  {
    try
    {
      LOCK.readLock().lock();
      int[] alignmentStartEnd = new int[] { 0, width - 1 };
      int startPos = alignmentStartEnd[0];
      int endPos = alignmentStartEnd[1];

      int[] lowestRange = new int[] { -1, -1 };
      int[] higestRange = new int[] { -1, -1 };

      if (hiddenColumns == null)
      {
        return new int[] { startPos, endPos };
      }

      for (int[] hiddenCol : hiddenColumns)
      {
        lowestRange = (hiddenCol[0] <= startPos) ? hiddenCol : lowestRange;
        higestRange = (hiddenCol[1] >= endPos) ? hiddenCol : higestRange;
      }

      if (lowestRange[0] == -1 && lowestRange[1] == -1)
      {
        startPos = alignmentStartEnd[0];
      }
      else
      {
        startPos = lowestRange[1] + 1;
      }

      if (higestRange[0] == -1 && higestRange[1] == -1)
      {
        endPos = alignmentStartEnd[1];
      }
      else
      {
        endPos = higestRange[0] - 1;
      }
      return new int[] { startPos, endPos };
    } finally
    {
      LOCK.readLock().unlock();
    }

  }

  /**
   * Finds the hidden region (if any) which starts or ends at res
   * 
   * @param res
   *          visible residue position, unadjusted for hidden columns
   * @return region as [start,end] or null if no matching region is found
   */
  public int[] getRegionWithEdgeAtRes(int res)
  {
    try
    {
      LOCK.readLock().lock();
      int adjres = adjustForHiddenColumns(res);

      int[] reveal = null;
      if (hiddenColumns != null)
      {
        for (int[] region : hiddenColumns)
        {
          if (adjres + 1 == region[0] || adjres - 1 == region[1])
          {
            reveal = region;
            break;
          }
        }
      }
      return reveal;
    } finally
    {
      LOCK.readLock().unlock();
    }
  }

}
